//***************************************************************
// 3DPlugin.cpp
// 
// Direct 3D Plugin Sample
// Andrew Skowronski, Copywrite 1999
// andrew_skow@hotmail.com
// 
// This sample creates a "fly over" of a 3D landscape
// that is generated based on the music coming from Winamp.
// It demonstrates Direct 3D Retained Mode
//
// Feel free to base your own plugins on this code.
//
//***************************************************************

#include "3DPlugin.h" // Defines C3DPlugin

#define D3D_OVERLOADS

#include <windows.h>
#include <ddraw.h>


// Global instance of the Plugin object
C3DPlugin    g_Plugin ;

// Trace file
CTraceOutput g_Info ;


// TODO: Make member of C3DPlugin?
extern HWND hMainWnd ; 

C3DPlugin::C3DPlugin()
	: m_pD3DRM( NULL )
	, m_pDDClipper( NULL )
	, m_pRMDevice( NULL )
	, m_pView( NULL )
	, m_pScene( NULL )
	, m_pCamera( NULL )
	, m_pFrontSurface( NULL )
	, m_pBackSurface( NULL )
	, m_bMinimized( FALSE )
	, m_bInitialized( FALSE )
	, m_pLandscapeMB( NULL )
	, m_pVertices( NULL )
	, m_iVertices( 0 )
	, m_pLandscapeFrame( NULL )
	, m_iTextureCnt( 0 )
	, m_pDD( NULL )
	, m_bIgnoreResize( FALSE )
	, m_bFirstInit( FALSE )
    , m_dwOldWidth( 640 )
    , m_dwOldHeight( 480 )
    , m_dwOldBPP( 16 )

{
}

void C3DPlugin::ResizeViewport( struct winampVisModule *in_mod )
{
	if ( !m_bFirstInit )
	{
		// This resize is occuring before we have been properly initialized
		return ;
	}

	if ( m_bIgnoreResize )
	{
		// Resize can be launched by Initialization operation - we need to test
		// so we don't reset
		return ;
	}

	TRACE0( "Resizing Viewport" ) ;

	// Remember the state...

	D3DVECTOR l_CameraPosition ;
	D3DVECTOR l_LandOrientUp ;
	D3DVECTOR l_LandOrientDir ;

	if ( ( m_pCamera == NULL ) ||
	     ( m_pLandscapeFrame == NULL ) )
	{
		// Unexpected
		return ;
	}
	m_pCamera->GetPosition( m_pScene, &l_CameraPosition ) ;

	m_pLandscapeFrame->GetOrientation( m_pScene, &l_LandOrientUp, &l_LandOrientDir ) ;

	// Do a complete cleanup and restore.  We might be able
	// to get away with less of a clean up, but it seems from
	// the examples that we completely have to re-create all the
	// basic elements.

	Quit( in_mod ) ;
	Init( in_mod ) ;

	// Restore the state
	// TODO: Should also restore the texture.

	m_pCamera->SetPosition(m_pScene, l_CameraPosition.x, l_CameraPosition.y, l_CameraPosition.z ) ;
	m_pLandscapeFrame->SetOrientation( m_pScene, 
									   l_LandOrientUp.x, l_LandOrientUp.y, l_LandOrientUp.z,
									   l_LandOrientDir.x, l_LandOrientDir.y, l_LandOrientDir.z ) ;

}

// Initialize C3DPlugin - Create D3D objects
// and create scene
// returns 1 in case of of failure
int C3DPlugin::Init( struct winampVisModule *in_mod )
{
	HRESULT hr = S_OK ;
	g_Info.Trace( "Initializing Plugin\n" ) ;

	m_bIgnoreResize = TRUE ;

	if ( m_bInitialized == TRUE )
	{
		// Something is wrong, we are going to leak if we
		// proceed
		ERRORMSG( "Initialized called twice?" ) ;
		return 1 ;
	}

	GetPluginConfig( in_mod, &m_Config ) ;

    int width;    // Device's width 
    int height;   // Device's height 

	///////////////////////////////////////////////////////////////////////////
    // Create a DirectDrawClipper object and associate the window with it.
	///////////////////////////////////////////////////////////////////////////

	hr = DirectDrawCreate( NULL, &m_pDD, NULL ) ;
    if (FAILED(hr)) 
    {
        MessageBox(in_mod->hwndParent, 
			       "Failed to create Direct Draw - Do you have the latest version of Direct X installed?", "Plugin Error", MB_OK ) ;
        return 1;
    }

	///////////////////////////////////////////////////////////////////////////
    // Create the Direct3DRM object.
    ///////////////////////////////////////////////////////////////////////////

    LPDIRECT3DRM pD3DRMTemp;
    hr = Direct3DRMCreate(&pD3DRMTemp);
    if (FAILED(hr)) 
    {
        MessageBox(in_mod->hwndParent, 
			       "Failed creation of Direct3DRM - Do you have the latest version of Direct X installed?", "Error", MB_OK ) ;
        return 1;
    }

    if( FAILED( pD3DRMTemp->QueryInterface(IID_IDirect3DRM3, (void **)&m_pD3DRM) ) )
    {
        RELEASE(pD3DRMTemp);
        MessageBox(in_mod->hwndParent, 
			       "Failed query for Direct3DRM3 - Do you have the latest version of Direct X installed?", "Error", MB_OK ) ;
        return 1;
    }
    RELEASE(pD3DRMTemp);

	if ( m_Config.bFullScreen )
	{
		hr = InitFullScreen() ;
	}
	else
	{
		hr = InitWindowVersion() ;
	}

	if ( FAILED( hr ) )
	{
		return 1 ;
	}

	if ( m_Config.iQuality > 0 )
	{
		hr = m_pRMDevice->SetQuality( D3DRMRENDER_GOURAUD ) ;
	}

    // Create the master scene frame and the camera frame.
    hr = m_pD3DRM->CreateFrame(NULL, &m_pScene);
    if (FAILED(hr)) 
    {
        ERRORMSG("Failed to create the master scene frame.");
        return FALSE;
    }

    hr = m_pD3DRM->CreateFrame(m_pScene, &m_pCamera);
    if (FAILED(hr)) 
    {
        ERRORMSG("Failed to create the camera's frame.");
        return FALSE;
    }

    // Create the Direct3DRM viewport using the device, camera frame,
    // and the device's width and height.
    width = m_pRMDevice->GetWidth();
    height = m_pRMDevice->GetHeight();

    hr = m_pD3DRM->CreateViewport(m_pRMDevice, m_pCamera, 0, 0, 
                                   width, height, &m_pView);
    if (FAILED(hr)) 
    {
		Quit( in_mod ) ;
        return FALSE;
    }

    hr = m_pD3DRM->CreateMeshBuilder( &m_pLandscapeMB ) ;

    if (FAILED(hr)) 
    {
		Quit( in_mod ) ;
        return FALSE;
    }

	m_bInitialized = TRUE ;
	m_bIgnoreResize = FALSE ;

	TRACE0( "Direct 3D Environment created" ) ;

	CreateScene( ) ;

	return 0 ;
}

HRESULT C3DPlugin::InitWindowVersion()
{
    RECT rc;      // Bounding rectangle for main window 
	LPDIRECTDRAW2 l_pDD2 = NULL ;
	HRESULT hr ;

	hr = m_pDD->QueryInterface( IID_IDirectDraw2, (LPVOID *)&l_pDD2 ) ;
	if ( FAILED( hr ) )
	{
		ERRORMSG( "Could not get Direct Draw 2 interface" ) ;
		return hr ;
	}

	hr = l_pDD2->CreateClipper( 0, &m_pDDClipper, NULL ) ;
	if ( FAILED ( hr ) )
	{
		ERRORMSG( "Could not create Clipper" ) ;
		return hr ;
	}

	RELEASE( l_pDD2 ) ;

    hr = m_pDDClipper->SetHWnd( 0, hMainWnd );
    if (FAILED(hr)) 
    {
        ERRORMSG("Failed to set the window handle for the DirectDrawClipper.");
        return hr;
    }

    // Create a default Direct3DRM device.
    GetClientRect(hMainWnd, &rc);

    hr = m_pD3DRM->CreateDeviceFromClipper(m_pDDClipper, 
                           &m_Config.guidDevice,
                           rc.right, rc.bottom, &m_pRMDevice);
   
	if (FAILED(hr)) 
    {
        ERRORMSG("Failed to create the preferred device.") ;
		
		if ( m_Config.guidDevice != GUID_NULL )
		{	
			hr = m_pD3DRM->CreateDeviceFromClipper(m_pDDClipper, 
                           NULL, // Default device 
                           rc.right, rc.bottom, &m_pRMDevice) ;
		}
        
		if ( FAILED ( hr ) )
		{
			ERRORMSG("Failed to create the default D3DRM device.");
			return E_FAIL ;
		}
    }

	TRACE0( "Direct 3D Device Created" ) ;

	return S_OK ;
}

// Deal with full screen mode working
// Its a royal pain that you have to do all this stuff
// to get it running, basically you have to set up the Direct Draw
// surfaces in Video memory yourself and then create the
// 3D device on that surface.
// Note: This code should have more ERRORMSG handling.

HRESULT C3DPlugin::InitFullScreen()
{
	HRESULT hr = S_OK ;
	BOOL bHardware = TRUE ;
	BOOL bBackBufferInVideo = TRUE ;

	hr = m_pDD->SetCooperativeLevel( hMainWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWMODEX );

	if ( FAILED ( hr ) )
	{
		MessageBox( hMainWnd, "Wouldn't Go to Exclusive", "Error", MB_OK ) ;
	}

	// From this point on we can't use message boxes any more, the
	// UI is frozen by Direct Draw....

	////////////////////////////////////////////////////////////////////////////
	// Remember current display settings so that we can recover them
	////////////////////////////////////////////////////////////////////////////
    DDSURFACEDESC ddsd;

    memset(&ddsd, 0, sizeof(DDSURFACEDESC));
    ddsd.dwSize = sizeof(DDSURFACEDESC);

	hr = m_pDD->GetDisplayMode( &ddsd ) ;

    m_dwOldWidth = ddsd.dwWidth;
    m_dwOldHeight = ddsd.dwHeight;
    m_dwOldBPP = ddsd.ddpfPixelFormat.dwRGBBitCount;

	hr = m_pDD->SetDisplayMode( FULL_SCREEN_WIDTH, FULL_SCREEN_HEIGHT, FULL_SCREEN_BPP ) ;
	if ( FAILED ( hr ) )
	{
		ERRORMSG( "Could not set Direct Draw display mode" ) ;
	}

	DDSCAPS ddscaps;

    memset(&ddsd, 0, sizeof(DDSURFACEDESC));
    ddsd.dwSize = sizeof( ddsd );
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP |
                          DDSCAPS_3DDEVICE | DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 1;

    if ( bHardware )
	    ddsd.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY;

	// Create the surface
    hr = m_pDD->CreateSurface( &ddsd, &m_pFrontSurface, NULL ) ;

	if ( FAILED ( hr ) )
	{
		ERRORMSG( "Could not create surface" ) ;
	}

	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
    hr = m_pFrontSurface->GetAttachedSurface( &ddscaps, &m_pBackSurface ) ;

	if ( FAILED( hr ) )
	{
		ERRORMSG( "Could not get back buffer" ) ;
		return hr ;
	}

	bBackBufferInVideo = (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) ? TRUE : FALSE ;

	if ( !bBackBufferInVideo )
	{
		ERRORMSG( "Back buffer is not in Video Memory" ) ;
	}

	// At this point we have Direct Draw Surfaces that we could
	// use for 2D effects.  In that case we would want to also
	// adapt the Windowed version to do this same thing, plus use
	// a Clipper object

	// We are going to render into the back buffer

    hr = m_pD3DRM->CreateDeviceFromSurface( &m_Config.guidDevice, m_pDD, m_pBackSurface, 0, &m_pRMDevice ) ;

	if ( FAILED( hr ) )
	{
		ERRORMSG( "Could not Create Device from Surface" ) ;
		return hr ;
	}

	// Clear the Direct Draw surfaces
	ClearSurface( m_pFrontSurface ) ;
	ClearSurface( m_pBackSurface ) ;

	TRACE0( "Direct 3D Device Created" ) ;

	return hr ;
}

// Fill a surfaces with black so that we don't have crap displayed
// for the first few frames....

void C3DPlugin::ClearSurface( LPDIRECTDRAWSURFACE in_pSurface )
{
	DDSURFACEDESC ddsd;
    RECT dst;
    DDBLTFX ddbltfx;
	HRESULT hr = S_OK ;

	if ( in_pSurface == NULL )
	{
		return ;
	}
	
    ZeroMemory( &ddsd, sizeof(ddsd) ) ;
    ddsd.dwSize = sizeof( ddsd );

    hr = in_pSurface->GetSurfaceDesc( &ddsd );

	if ( FAILED ( hr ) )
	{
		return ;
	}
	
	memset(&ddbltfx, 0, sizeof(ddbltfx)) ;
    ddbltfx.dwSize = sizeof(DDBLTFX) ;

    SetRect(&dst, 0, 0, ddsd.dwWidth, ddsd.dwHeight);
    hr = in_pSurface->Blt( &dst, 
		                   NULL, NULL, 
						   DDBLT_COLORFILL | DDBLT_WAIT,
						   &ddbltfx ) ;
}

// Creates the lighting for the project
HRESULT C3DPlugin::InitializeLights()
{
	HRESULT hr = S_OK ;
	
	// DEFAULTS:								//Bright // Position        //Rot		  //Amt
	m_aLights[0].Init( D3DRMLIGHT_SPOT,			 0.6f, 0.0f, 4.0f, 0.0f, 0.9f, 0.0f, 0.0f, 0.1f, m_pScene, m_pD3DRM ) ;
	m_aLights[1].Init( D3DRMLIGHT_SPOT,			 0.9f, 5.0f, 5.0f, 5.0f,-1.0f, 0.2f, 0.2f, 0.05f,m_pScene, m_pD3DRM ) ;
	m_aLights[2].Init( D3DRMLIGHT_PARALLELPOINT, 0.6f,-2.0f,10.0f,-5.0f, 0.0f, 0.0f, 0.0f, 0.0f, m_pScene, m_pD3DRM ) ;		
	
	// Extras:

	// Ambient light
	m_aLights[3].Init( D3DRMLIGHT_AMBIENT,       0.4f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, m_pScene, m_pD3DRM ) ;

	// Underneath light
	m_aLights[4].Init( D3DRMLIGHT_PARALLELPOINT, 0.7f, 8.0f,-5.0f,-8.0f, 0.0f, 0.0f, 0.0f, 0.0f, m_pScene, m_pD3DRM ) ;

	// Fast Spot Lights
	m_aLights[5].Init( D3DRMLIGHT_SPOT,			 0.9f, 5.0f, 5.0f, 5.0f,-1.0f,-0.2f, 0.5f, 0.5f, m_pScene, m_pD3DRM ) ;
	m_aLights[6].Init( D3DRMLIGHT_SPOT,			 0.5f, 0.0f, 1.0f, 0.3f, 0.0f, 0.9f,-0.1f, 0.4f, m_pScene, m_pD3DRM ) ;

	// Brighter Main Light
	m_aLights[7].Init( D3DRMLIGHT_PARALLELPOINT, 1.0f,-2.0f,10.0f,-5.0f, 0.0f, 0.0f, 0.0f, 0.0f, m_pScene, m_pD3DRM ) ;

	// Turn on the default lights
	for ( int i = 0 ; i < DEFAULT_LIGHTS ; i++ )
	{
		m_aLights[i].ToggleActive() ;
	}

	return hr ;
}

// Create the 3D objects for the scene
HRESULT C3DPlugin::CreateScene( )
{
    HRESULT hr = S_OK ;

	hr = AddFog() ;
	if ( FAILED( hr ) ) return hr ;

	hr = AddLandscape() ;
	if ( FAILED( hr ) ) return hr ;

    hr = m_pCamera->SetPosition(m_pScene, D3DVAL(CAMERA_X), D3DVAL(CAMERA_Y), D3DVAL(CAMERA_Z));
	if ( FAILED( hr ) ) return hr ;

	hr = InitializeLights() ;
	if ( FAILED( hr ) ) return hr ;

	TRACE0( "Direct 3D Scene Created" ) ;

    return 0;
}

// Create the Landscape and add it to the scene
HRESULT C3DPlugin::AddLandscape()
{
    LPDIRECT3DRMMATERIAL2 pMat            = NULL;

	HRESULT hr = S_OK ;

	double dZ ;
	double dX ;

	if ( m_Config.iQuality == 0 )
	{
		dZ = ( Z2 - Z1 ) / ( double ) Z_STEPS_LOW_RES ;
		dX = ( X2 - X1 ) / ( double ) X_STEPS_LOW_RES ;
	}
	else
	{
		dZ = ( Z2 - Z1 ) / ( double ) Z_STEPS ;
		dX = ( X2 - X1 ) / ( double ) X_STEPS ;
	}

	hr = CreateSurface( X1, 
						X2, 
						dX, 
						Z1, 
						Z2, 
						dZ, 
						LAND_Y_POS,		// Y Position
						m_iVertices,
						&m_pVertices,
						m_pLandscapeMB ) ;

    // Create a child frame within the scene.
    hr = m_pD3DRM->CreateFrame( m_pScene, &m_pLandscapeFrame ) ;

	if( FAILED( hr ) ) goto cleanup ;

	// Rotate the Landscape in perspective to the Scene
	// This was easier than rewriting the CreateScene method
	// or moving the camera
	hr = m_pLandscapeFrame->SetOrientation( m_pScene, 
										  D3DVAL(1), D3DVAL(0), D3DVAL( 0 ), 
										  D3DVAL(0), D3DVAL(1), D3DVAL( 0 ) ) ;

	if( FAILED( hr ) ) goto cleanup ;

    // Add the loaded mesh into the child frame.
    hr = m_pLandscapeFrame->AddVisual( (LPDIRECT3DRMVISUAL)m_pLandscapeMB ) ;

    // Create a material, setting the reflectiveness (5.0 is metallic, 
    // higher is more plastic)
    hr = m_pD3DRM->CreateMaterial( D3DVAL( LAND_MATERIAL ), &pMat ) ;

	if( FAILED( hr ) ) goto cleanup ;
    
    hr = m_pLandscapeMB->SetMaterial(pMat) ;

	if( FAILED( hr ) ) goto cleanup ;

    // Set the landscape color
    hr = m_pLandscapeMB->SetColor( m_Config.ObjectColor ) ;

	if( FAILED( hr ) ) goto cleanup ;

	// We ignore error return here, program
	// can run without textures.
//	InitTexture() ;

cleanup:

	RELEASE( pMat ) ;

	return hr ;
}

// Creates a Grid of Polygons in the X, Z Plane at Y = y
// Plane is from (in_x1 to in_x2) on X-axis and (in_z1 to in_z2) on Z-axis
// Provided Mesh builder is filled with the surface
// and vertices returned.  Caller responsible to delete

HRESULT C3DPlugin::CreateSurface( double in_x1, double in_x2, double in_dx,
                                  double in_z1, double in_z2, double in_dz,
								  double in_y, 
								  int & out_iVertices, 
								  D3DVECTOR ** out_ppVertices,
								  LPDIRECT3DRMMESHBUILDER3 in_pMB )
{
	HRESULT hr = S_OK ;
    
	// Determine number of vertices
	int iRowVertices = (int)((in_x2 - in_x1) / in_dx) ;
    int iColumnVertices = (int)((in_z2 - in_z1) / in_dz) ;
    if ((iRowVertices < 1) || (iColumnVertices < 1)) return FALSE ;
    out_iVertices = (iRowVertices + 1) * (iColumnVertices + 1);

    // Allocate Vertices

	*out_ppVertices = new D3DVECTOR [ out_iVertices] ;
    D3DVECTOR* pv = m_pVertices;

    // Create the array for the face data
    // each face will have 4 vertices
    int iFaces = iRowVertices * iColumnVertices;
    int* FaceData = new int [iFaces * 5 + 1];
    int* pfd = FaceData;

    // write out the vertex set
    double x = in_x1;
    double z;
    for (int iRow = 0; iRow <= iRowVertices; iRow++) 
	{
        z = in_z1;
        for (int iCol = 0; iCol <= iColumnVertices; iCol++) 
		{
            pv->x = D3DVAL(x);
            pv->z = D3DVAL(z);
            pv->y = D3DVAL( in_y ) ;
            pv++;
            z += in_dz;
        }
        x += in_dx;
    }

    // write out the face list
    int iFirst = iColumnVertices + 1;
    for (iRow = 0; iRow < iRowVertices; iRow++) 
	{
        for (int iCol = 0; iCol < iColumnVertices; iCol++) 
		{
            *pfd++ = 4; // Polygon has four vertices
            *pfd++ = iFirst + iCol;
            *pfd++ = iFirst - iColumnVertices - 1 + iCol;
            *pfd++ = iFirst - iColumnVertices + iCol;
            *pfd++ = iFirst + iCol + 1;
        }
        iFirst += iColumnVertices + 1;
    }
    *pfd = 0; // end the list

    // Create the surface with auto-generation of the normals

	hr = in_pMB->AddFaces(out_iVertices, 
						  *out_ppVertices, 
						  0, 
						  NULL, 
						  (ULONG*) FaceData, 
						  NULL ) ;

	hr = in_pMB->GenerateNormals(0.0, 0) ;
	
	// Cleanup
	delete [] FaceData;

	return hr ;
}

HRESULT C3DPlugin::AddFog()
{
	HRESULT hr = S_OK ;

	// Maybe this will work....

//	hr = m_pRMDevice->SetDither( TRUE ) ;

#if 0
	hr = m_pScene->SetFogEnable( TRUE ) ;

	hr = m_pScene->SetFogParams( 1, 30, 1 ) ;
#endif

	//
	// I can't get fog working yet.  Do I have to set something
	// else on the Device or Scene?
	//

	hr = m_pScene->SetSceneFogEnable( TRUE ) ;

	hr = m_pScene->SetSceneFogMethod( D3DRMFOGMETHOD_VERTEX ) ;

	hr = m_pScene->SetSceneFogMode( D3DRMFOG_LINEAR ) ;

	hr = m_pScene->SetSceneFogParams( D3DVALUE( 1 ), D3DVALUE( 5 ), D3DVALUE( 0 ) ) ;

	D3DCOLOR FogColor = m_Config.ObjectColor ;

	hr = m_pScene->SetSceneFogColor( FogColor ) ;



	return hr ;
}


// Change the landscape for one render clock tick
HRESULT C3DPlugin::UpdateLandscape( struct winampVisModule *in_mod )
{
	HRESULT hr ;

	double dZ ;
	double dX ;

	if ( m_Config.iQuality == 0 )
	{
		dZ = ( Z2 - Z1 ) / ( double ) Z_STEPS_LOW_RES ;
		dX = ( X2 - X1 ) / ( double ) X_STEPS_LOW_RES ;
	}
	else
	{
		dZ = ( Z2 - Z1 ) / ( double ) Z_STEPS ;
		dX = ( X2 - X1 ) / ( double ) X_STEPS ;
	}

	int nColumnVert = ( int) ( ( Z2 - Z1 ) / dZ ) + 1;
	int nRowVert = ( int) ( ( X2 - X1 ) / dX ) + 1;
	int nVert = nColumnVert * nRowVert ;

	// Slide the vertices down a row

	D3DVECTOR* pDest = &m_pVertices[ nColumnVert ] ;

	memmove( pDest, 
			 m_pVertices, 
			 ( nVert - nColumnVert ) * sizeof( D3DVECTOR ) ) ;

	// Sliding down requires that we update all the X positions
	// of each vertice

	for ( int j = 0 ; j < ( nVert - nColumnVert ) ; j++ )
	{
		pDest[ j ].x += D3DVAL( dX ) ;
	}

	// Use the data from Winamp to draw the new row of
	// landscape
	// You can screw around a lot with this to get different
	// looking landscapes

	double z = Z1 ;
	z += dZ ;

	float fRough = ( float ) ( LANDSCAPE_ROUGH_MAX - m_Config.iLandscapeRough + LANDSCAPE_ROUGH_MIN ) ;

	for ( int i = 1 ; i < ( nColumnVert - 1 ) / 2 ; i++ )
	{
		unsigned char AudioData = in_mod->spectrumData[0][ i * AUDIO_BANDWIDTH_SPREAD ] ;

		float dy = LAND_Y_POS + // Base Position
			  	   ( ( float ) AudioData / fRough ) ;

		m_pVertices[ i ].x = D3DVAL( X1 ); // First Row
		m_pVertices[ i ].y = D3DVAL( dy ) ;
		m_pVertices[ i ].z = D3DVAL( z );

		z += dZ ;
	}

	for (  ; i < nColumnVert - 1 ; i++ )
	{
		// This is the second half of the new Row
		// Use the other channel of Audio and go in the
		// opposite direction

		unsigned char AudioData = in_mod->spectrumData[1][ ( nColumnVert - 1 - i ) * AUDIO_BANDWIDTH_SPREAD ] ;

		float dy = LAND_Y_POS + // Base Position
			  	   ( ( float ) AudioData / fRough  ) ;
		
		m_pVertices[ i ].x = D3DVAL( X1 ); // First Row
		m_pVertices[ i ].y = D3DVAL ( dy ) ;
		m_pVertices[ i ].z = D3DVAL( z );

		z += dZ ;
	}

	// Override the existing vertices

	hr = m_pLandscapeMB->SetVertices( 0, nVert, m_pVertices ) ;

	// Might be area to optimize for performance - is it necessary?

	hr = m_pLandscapeMB->GenerateNormals(0.0, 0) ;

	return hr ;
}

// Called constantly by Winamp to update the visualization
int C3DPlugin::Render(struct winampVisModule *in_mod)
{
	if ( !m_bInitialized && m_bMinimized ) 
	{
		// Nothing we can do
		return 0 ;
	}

	HRESULT hr = S_OK ;

	hr = UpdateLandscape( in_mod ) ;

	// If there is any rotation or animation:
	
    // Move the scene.
    hr = m_pScene->Move(D3DVAL(1.0));
    if (FAILED(hr)) 
    {
        ERRORMSG("Moving scene failed.");
        return FALSE;
    }

	// Clear the viewport.
    hr = m_pView->Clear(D3DRMCLEAR_ALL);
    if (FAILED(hr)) 
    {
        ERRORMSG("Clearing viewport failed.");
        return 1 ;
    }

    // Render the scene to the viewport
    hr = m_pView->Render( m_pScene );
    if (FAILED(hr)) 
    {
        ERRORMSG("Rendering scene failed.");
        return 1;
    }

    // Update the window.
    hr = m_pRMDevice->Update();
    if (FAILED(hr)) 
    {
        ERRORMSG("Updating device failed.");
        return 1;
    }

	if ( m_Config.bFullScreen )
	{
        hr = m_pFrontSurface->Flip( m_pBackSurface,
                                    1);
	}

	return 0 ;
}

// Clean up
void C3DPlugin::Quit(struct winampVisModule *in_mod)
{
	TRACE0( "Cleaning up Plugin" ) ;

	////////////////////////////////////////////////////////////////////////////
	// Recover the user's settings
	////////////////////////////////////////////////////////////////////////////
	if ( ( m_pDD != NULL ) && ( m_Config.bFullScreen ) )
	{
		TRACE0( "Restoring Display mode" ) ;
		m_pDD->SetDisplayMode( m_dwOldWidth, m_dwOldHeight, m_dwOldBPP ) ;

		TRACE0( "Releasing Exclusive Mode" ) ;
		m_pDD->SetCooperativeLevel( hMainWnd, DDSCL_NORMAL );
	}

	////////////////////////////////////////////////////////////////////////////
	// Let go of all our COM object references
	// We give a fair bit of debug information because its
	// pretty easy to have a crash here if you release out of order...
	////////////////////////////////////////////////////////////////////////////
	TRACE0( "Releasing COM objects" ) ;
	TRACE0( "----- 3D Scene" ) ;

	RELEASE( m_pLandscapeMB ) ;
	RELEASE( m_pLandscapeFrame ) ;
    RELEASE( m_pCamera );
    RELEASE( m_pScene );

	TRACE0( "----- 3D View" ) ;
    RELEASE( m_pView );

	TRACE0( "----- Clipper" ) ;
    RELEASE( m_pDDClipper);

	TRACE0( "----- 3D Device" ) ;
    RELEASE( m_pRMDevice );

	TRACE0( "----- Direct 3D Retained Mode" ) ;
    RELEASE( m_pD3DRM);

	TRACE0( "----- DD Surfaces" ) ;

	RELEASE( m_pBackSurface ) ;
	RELEASE( m_pFrontSurface ) ;

	TRACE0( "----- Direct Draw" ) ;
	RELEASE( m_pDD ) ;

	////////////////////////////////////////////////////////////////////////////
	// Free our allocated memory
	////////////////////////////////////////////////////////////////////////////

	TRACE0( "Freeing Allocated Memory" ) ;
    SAFE_DELETE( m_pVertices );

	m_bInitialized = FALSE ;
}


// Take care of a keyboard input
// Returns FALSE if the input key doesn't 
// mean anything to the plugin so that it can
// be processed by winamp

BOOL C3DPlugin::HandleKey( UINT in_Operation, WPARAM in_Key )
{
	BOOL bHandled = TRUE ;

	if ( m_pScene == NULL ) return FALSE ;
	if ( m_pCamera == NULL ) return FALSE ;

	if ( in_Operation == WM_KEYDOWN )
	{		
		D3DVECTOR dir, up, right ;

	    m_pCamera->GetOrientation( m_pScene, &dir, &up ) ;
	    D3DRMVectorCrossProduct( &right, &up, &dir );

		// Slow down a little bit
	    up.x /= D3DVAL(2.0);
	    up.y /= D3DVAL(2.0);
	    up.z /= D3DVAL(2.0);
	    right.x /= D3DVAL(2.0);
	    right.y /= D3DVAL(2.0);
	    right.z /= D3DVAL(2.0);

	    switch ( in_Key )
	    {
			case '1' :
			case '2' :
			case '3' :
			case '4' :
			case '5' :
			case '6' :
			case '7' :
			case '8' :				
			case '9' :				
				if ( in_Key - '0' <= NUM_LIGHTS )
				{
					m_aLights[ in_Key - '1' ].ToggleActive() ;
				}
				break ;

			case VK_INSERT:
				m_pLandscapeFrame->SetRotation(m_pScene, 
								  			   D3DVAL(1), D3DVAL(0), D3DVAL(0), 
											   D3DVAL(-0.1) ) ; // angle 
				break;

			case VK_DELETE:
				m_pLandscapeFrame->SetRotation(m_pScene, 
							  				   D3DVAL(1), D3DVAL(0), D3DVAL(0), 
											   D3DVAL(0.1) ) ; // angle 
				break;

			case VK_HOME:
				m_pLandscapeFrame->SetRotation(m_pScene, 
								  			   D3DVAL(0), D3DVAL(1), D3DVAL(0), 
											   D3DVAL(-0.1) ) ; // angle 
				break;

			case VK_END:
				m_pLandscapeFrame->SetRotation(m_pScene, 
							  				   D3DVAL(0), D3DVAL(1), D3DVAL(0), 
											   D3DVAL(0.1) ) ; // angle 
				break;

			case 'T':
				m_pCamera->SetVelocity(m_pScene, dir.x, dir.y, dir.z, FALSE);
				break;

		    case 'R':
				m_pCamera->SetVelocity(m_pScene, -dir.x, -dir.y, -dir.z, FALSE);
				break;

			case VK_NUMPAD8:
			case VK_UP:
				m_pCamera->SetVelocity(m_pScene, up.x, up.y, up.z, FALSE);
				break;

			case VK_NUMPAD2:
			case VK_DOWN:
				m_pCamera->SetVelocity(m_pScene, -up.x, -up.y, -up.z, FALSE);
				break;

			case VK_NUMPAD6:
			case VK_RIGHT:
				m_pCamera->SetVelocity(m_pScene, right.x, right.y, right.z, FALSE);
				break;

			case VK_NUMPAD4:
			case VK_LEFT:
				m_pCamera->SetVelocity(m_pScene, -right.x, -right.y, -right.z, FALSE);
				break;

			case VK_NUMPAD5:
				// Pop back to default position
				m_pCamera->SetVelocity(m_pScene, D3DVAL(0.0), D3DVAL(0.0), D3DVAL(0.0), FALSE );
				m_pCamera->SetPosition(m_pScene, D3DVAL(CAMERA_X), D3DVAL(CAMERA_Y), D3DVAL(CAMERA_Z) );

				m_pLandscapeFrame->SetOrientation( m_pScene, 
										  D3DVAL(1), D3DVAL(0), D3DVAL( 0 ), 
										  D3DVAL(0), D3DVAL(1), D3DVAL( 0 ) ) ;
				break;

			case VK_SPACE :
				ShowNextTexture() ;

			default:
				bHandled = FALSE ;
				break ;
		}
	}
	else if ( in_Operation ==  WM_KEYUP )
	{
		switch ( in_Key )
		{
			case 'T':
			case 'R':
			case VK_UP:
			case VK_DOWN:
			case VK_LEFT:
			case VK_RIGHT:
				// Stop the camera motion
				m_pCamera->SetVelocity( m_pScene, D3DVAL(0.0), D3DVAL(0.0), D3DVAL(0.0), FALSE ) ;
				break;

			case VK_INSERT:
			case VK_DELETE:
				// Stop the rotation
				m_pLandscapeFrame->SetRotation(m_pScene, 
							  				   D3DVAL(1), D3DVAL(0), D3DVAL(0), 
											   D3DVAL(0.0) ) ; // angle 
				break;

			case VK_HOME:
			case VK_END:
				// Stop the rotation
				m_pLandscapeFrame->SetRotation(m_pScene, 
							  				   D3DVAL(0), D3DVAL(1), D3DVAL(0), 
											   D3DVAL(0.0) ) ; // angle 
				break;

			default:
				bHandled = FALSE ;
				break ;
		}
	}
	else
	{
		bHandled = FALSE ;
	}

	return bHandled ;
}

// Cycle to the next texture
HRESULT C3DPlugin::ShowNextTexture()
{
    LPDIRECT3DRMWRAP     pWrap = NULL;
    LPDIRECT3DRMTEXTURE3 pTex = NULL;
	HRESULT hr ;

	if ( m_Config.szTextureFile[ 0 ] == 0 )
	{
		// No texture file defined
		return S_FALSE ;
	}

	char szFileToLoad[ MAX_PATH ] ;

	// We support series of files
	if ( m_iTextureCnt > 0 )
	{		
		// Try to load next file in the series
		char szFileDrive[ 3 ] ;
		char szFilePath[ MAX_PATH ] ;
		char szExt[ MAX_PATH ] ;
		char szFile[ MAX_PATH ] ;

		_splitpath( m_Config.szTextureFile, szFileDrive, szFilePath, szFile, szExt ) ;
		sprintf( szFileToLoad, "%s%s%s%d%s", szFileDrive, szFilePath, szFile, m_iTextureCnt + 1, szExt ) ;
	}
	else
	{
		// Try to load the first file
		strcpy( szFileToLoad, m_Config.szTextureFile ) ;
	}

	if ( GetFileAttributes( szFileToLoad ) == ( DWORD ) -1 )
	{
		hr = m_pLandscapeMB->SetTexture( NULL ) ;

		m_iTextureCnt = 0 ;

		ERRORMSG( "Texture file not found" ) ;

		return STG_E_FILENOTFOUND ;
	}

    hr = m_pD3DRM->LoadTexture( szFileToLoad, &pTex ) ;
    if (FAILED(hr)) 
    {
		ERRORMSG( "Texture file could not be loaded" ) ;
        return hr ;
    }

	hr = m_pLandscapeMB->SetTexture( pTex ) ;
	if( FAILED( hr ) )
	{
		ERRORMSG( "Set Texture Failed" ) ;	
		goto cleanup ;
	}

	// The arguments to this one are a real nightmare.

    hr = m_pD3DRM->CreateWrap( D3DRMWRAP_FLAT, NULL, 
							   D3DVAL( X1 ), D3DVAL( LAND_Y_POS ), D3DVAL( Z1 ),
							   0.0f, 1.0f, 0.0f, 
							   1.0f, 0.0f, 0.0f, 
							   0.0f, 0.0f,
							   D3DDivide( 1.0 * m_Config.iTileCnt, ( Z2 - Z1 ) ), 
							   D3DDivide( 1.0 * m_Config.iTileCnt, ( X2 - X1 ) ),
							   &pWrap ) ;
	if( FAILED( hr ) )
	{
		ERRORMSG( "Create Wrap Failed" ) ;	
		goto cleanup ;
	}

    hr = pWrap->Apply( (LPDIRECT3DRMOBJECT) m_pLandscapeMB ) ;
	if( FAILED( hr ) )
	{
		ERRORMSG( "Failed to apply wrap to Landscape" ) ;
		goto cleanup ;
	}

	m_iTextureCnt++ ;

cleanup:
	RELEASE( pTex ) ;
	RELEASE( pWrap ) ;

	return hr ;
}

void C3DPlugin::HandleActivate( WPARAM in_wparam )
{
	LPDIRECT3DRMWINDEVICE pWinDevice = NULL ;

	if ( m_pRMDevice == NULL )
	{
		return ;		
	}

	HRESULT hr = m_pRMDevice->QueryInterface( IID_IDirect3DRMWinDevice, (LPVOID*) &pWinDevice ) ;

	if ( SUCCEEDED( hr ) )
	{
		pWinDevice->HandleActivate( in_wparam ) ;
		RELEASE( pWinDevice ) ;
	}
}

void C3DPlugin::HandlePaint( HDC in_hdc )
{
	LPDIRECT3DRMWINDEVICE pWinDevice = NULL ;

	if ( m_pRMDevice == NULL )
	{
		return ;		
	}

	HRESULT hr = m_pRMDevice->QueryInterface( IID_IDirect3DRMWinDevice, (LPVOID*) &pWinDevice ) ;

	if ( SUCCEEDED( hr ) )
	{
		pWinDevice->HandlePaint( in_hdc ) ;
		RELEASE( pWinDevice ) ;
	}
}


