//******************************************************************************
// 3DPlugin.h
// Main plug in
//
// Andrew Skowronski, Copywrite 1999
// andrew_skow@hotmail.com
// 
// A good start to playing with this code is to change the
// DEFINE values and seeing how it changes the visualization
//******************************************************************************

#ifndef _VIS_D3D_H
#define _VIS_D3D_H

#include <d3drm.h>		// D3D retained Mode
#include <d3drmwin.h>
#include <stdio.h>

#include "vis.h"		// Winamp SDK Header File

#include "configdlg.h"	// Helper Class Header files
#include "light.h"

//******************************************************************************
// Configuration Defines
//
// These values define the behavior of the plug in.
// Some of these values are candidates for an enhanced
// configuration dialog.
//
//******************************************************************************

// Change the C3DPlugin::InitializeLights function
// if you change these
#define NUM_LIGHTS     8
#define DEFAULT_LIGHTS 3

//
// Defines the default Camera Position
//

#define CAMERA_X 0
#define CAMERA_Y 0
#define CAMERA_Z -9

//
// Window size
//
#define WINDOW_SIZE_WIDTH  600
#define WIDNOW_SIZE_HEIGHT 450

//
// Controls how often UI is refreshed and hence speed
//
#define LATENCY				20
#define DELAY				40

// Defines the coordinates and granularity of the
// Landscape in the Z direction and X direction

#define Z_STEPS			40 // Should be round number
#define Z_STEPS_LOW_RES 10

#define Z1 -10.0
#define Z2 10.0

#define X_STEPS			30 // Should be round number
#define X_STEPS_LOW_RES 10

#define X1 -10.0
#define X2 10.0

// Landscape is based on the plane y = LAND_Y_POS
#define LAND_Y_POS -2.0f

#define LANDSCAPE_ROUGH 250    // Larger value means less rough

#define LANDSCAPE_ROUGH_MIN 5
#define LANDSCAPE_ROUGH_MAX 500

// We get audio data spread across 576 positions
// For techno the bottom area tends to be most interesting
// A smaller value of AUDIO_BANDWIDTH_SPREAD means that we 
// are concentrated in that area.

#define AUDIO_BANDWIDTH_SPREAD 3


// Landscape Color

#define LAND_COLOR_R 0.2
#define LAND_COLOR_G 0.2
#define LAND_COLOR_B 1.0


// Material of the Landscape ( 5.0 Plastic, Large is metallic )
#define LAND_MATERIAL 5

// Useful for Full screen mode, which can freeze the UI when running,
// when this is defined the application will automatically kill itself after
// the timeout period
//#define ESCAPE_CRAP_OUT
//#define TIMER_RATE          30000	// Milliseconds

// Ideally we let the user choose the screen format for
// the full screen mode.  This is a safe choice, but
// the user might be able to run something much better....

#define FULL_SCREEN_WIDTH 640
#define FULL_SCREEN_HEIGHT 480
#define FULL_SCREEN_BPP 16

//******************************************************************************
// Plugin Configuration Structure
//******************************************************************************

typedef struct _PluginConfig
{
	int			iQuality ; // 0 is low quality...
	D3DCOLOR    ObjectColor ;
	char        szTextureFile[ MAX_PATH ] ;
	GUID        guidDevice ;
	BOOL        bFullScreen ;
	int         iResX ;        // Window size or Full Screen Res
	int         iResY ;
	int         iTileCnt ;     // Number of times to tile the texture
	int         iLandscapeRough ;
} PluginConfig, * LPPLUGINCONFIG ;

void GetDefaultPluginConfig( LPPLUGINCONFIG out_Config ) ;
void GetPluginConfig( struct winampVisModule * in_mod, LPPLUGINCONFIG out_Config ) ;
void SavePluginConfig( struct winampVisModule * in_mod, LPPLUGINCONFIG in_Config ) ;

//******************************************************************************
// C3DPlugin - Main Implementation class of the Plugin
//******************************************************************************

class CTraceOutput
{
	FILE * m_pFile ;
public :	
	CTraceOutput() ;
	~CTraceOutput() ;
	void Init() ;
	void Term() ;	
	void Trace( char * in_pszMsg ) ;
	void ErrorTrace( char * in_pszMsg ) ;
} ;

#define USE_ERROR_TRACING  extern CTraceOutput g_Info ;
#define TRACE0( x ) g_Info.Trace( x )
#define ERRORMSG( x )  g_Info.ErrorTrace( x )

//******************************************************************************
// C3DPlugin - Main Implementation class of the Plugin
//******************************************************************************

class C3DPlugin
{
public:
	C3DPlugin() ;
	int Init( struct winampVisModule *this_mod ) ;
	void Quit(struct winampVisModule *this_mod);   // deinitialization for module
	int Render(struct winampVisModule *this_mod) ;
	void ResizeViewport( struct winampVisModule *this_mod ) ;

	BOOL HandleKey( UINT in_Operation, WPARAM in_Key ) ;
	void HandleActivate(WPARAM) ;
	void HandlePaint(HDC) ;

private:
	LPDIRECTDRAW        m_pDD ;
	LPDIRECT3DRM3       m_pD3DRM ;     // Direct3DRM object 
	LPDIRECTDRAWCLIPPER m_pDDClipper ; // DirectDrawClipper object 

    // Direct3DRM device. 
    LPDIRECT3DRMDEVICE3   m_pRMDevice ;           

    // Direct3DRM viewport through which to view the scene. 
    LPDIRECT3DRMVIEWPORT2 m_pView;  

    // Master frame in which other frames are placed. 
    LPDIRECT3DRMFRAME3 m_pScene;    

    // Frame describing the users point of view. 
    LPDIRECT3DRMFRAME3 m_pCamera;   

	LPDIRECT3DRMFRAME3 m_pLandscapeFrame ;

	LPDIRECTDRAWSURFACE m_pFrontSurface ;
	LPDIRECTDRAWSURFACE m_pBackSurface ;

	// Landscape is defined by the following:
    LPDIRECT3DRMMESHBUILDER3 m_pLandscapeMB ;
	D3DVECTOR*				 m_pVertices ;
	int                      m_iVertices ;

	CLight					 m_aLights[ NUM_LIGHTS ] ;

	PluginConfig			 m_Config ;
	int						 m_iTextureCnt ;

	// We remember the display settings so that
	// we can recover them
	DWORD					m_dwOldBPP;
	DWORD					m_dwOldHeight;
	DWORD				    m_dwOldWidth;

public:
    // All D3DRM objects have been initialized. 
    BOOL m_bInitialized;          
    BOOL m_bMinimized;            
	BOOL m_bIgnoreResize ;
	BOOL m_bFirstInit ;

private:
	void    ClearSurface( LPDIRECTDRAWSURFACE in_pSurface ) ;
	HRESULT InitFullScreen() ;
	HRESULT InitWindowVersion() ;
	HRESULT ShowNextTexture();
	HRESULT CreateScene( ) ;
	HRESULT AddLandscape( ) ;
	HRESULT InitializeLights() ;
	HRESULT AddFog( ) ;
	HRESULT CreateSurface(double x1, double x2, double dx,
                          double z1, double z2, double dz,
						  double y,
						  int &, 
						  D3DVECTOR **,
						  LPDIRECT3DRMMESHBUILDER3 in_pMB ) ;
	HRESULT UpdateLandscape( struct winampVisModule *this_mod ) ;
} ;

//******************************************************************************
// Helpful Macros
//
// Safe cleanup MACROs
//******************************************************************************

#define RELEASE(x) if (x != NULL) {x->Release(); x = NULL;} 
#define SAFE_DELETE(x) if (x != NULL) { delete x; x = NULL;} 

//******************************************************************************
// TODO List:
//
// Call IDirect3DRMWinDevice::HandlePaint etc
//******************************************************************************
#endif