//******************************************************************************
// vis_d3d.cpp
// 
// Direct 3D Plugin Sample
// Andrew Skowronski, andrew_skow@hotmail.com
// This code takes care of creating a window, talking
// to winamp.  The actual 3D visualization is implemented
// in the C3DPlugin object.
//
// Based on:Winamp test visualization library v1.0, Copyright (C) 1997, Justin Frankel/Nullsoft
// 
// Feel free to base any plugins on this code.
//
//******************************************************************************

#include <windows.h>

#include "3DPlugin.h"   // Plug Header File

#define TIMER_ID            1


char szAppName[] = "D3DVis"; // Our window class, etc

extern C3DPlugin g_Plugin ;
USE_ERROR_TRACING

// configuration declarations
int config_x=50, config_y=50;	// screen X position and Y position, repsectively
void config_read(struct winampVisModule *this_mod);		// reads the configuration
void config_write(struct winampVisModule *this_mod);	// writes the configuration
void config_getinifn(struct winampVisModule *this_mod, char *ini_file); // makes the .ini file filename


// returns a winampVisModule when requested. Used in hdr, below
winampVisModule *getModule(int which);

// "member" functions
void config(struct winampVisModule *this_mod); // configuration dialog
int initD3D(struct winampVisModule *this_mod);	   // initialization for module
int renderD3D(struct winampVisModule *this_mod);  // rendering for module 2
void quitD3D(struct winampVisModule *this_mod);   // deinitialization for module

// our window procedure
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
HWND hMainWnd; // main window handle

// Double buffering data
HDC memDC;		// memory device context
HBITMAP	memBM;  // memory bitmap (for memDC)
HBITMAP oldBM;  // old bitmap (from memDC)


// Module header, includes version, description, and address of the module retriever function
winampVisHeader hdr = { VIS_HDRVER, "Direct 3D Sample Visualization v1.0", getModule };

// 
// This structure is used to communicate back and forth
// with winamp.
//
winampVisModule mod =
{
	"3D Landscape",
	NULL,	// hwndParent
	NULL,	// hDllInstance
	0,		// sRate
	0,		// nCh
	LATENCY,// latencyMS - How long to draw?
	DELAY,  // delayMS
	2,		// spectrumNch ( we want stereo data )
	0,		// waveformNch
	{ 0, },	// spectrumData (Spectrum will be filled into this)
	{ 0, },	// waveformData
	config,	// Callback functions for Winamp to Call
	initD3D,
	renderD3D, 
	quitD3D
};

// this is the only exported symbol. returns our main header.
// if you are compiling C++, the extern "C" { is necessary, so we just #ifdef it
#ifdef __cplusplus
extern "C" {
#endif
__declspec( dllexport ) winampVisHeader *winampVisGetHeader()
{
	return &hdr;
}
#ifdef __cplusplus
}
#endif

// getmodule routine from the main header. Returns NULL if an invalid module was requested.
// Winamp will use this to enumerate through the available modules.
winampVisModule *getModule(int which)
{
	switch (which)
	{
		case 0: return &mod;
		default:return NULL;
	}
}

// window procedure for our window
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	
	switch (message)
	{
		case WM_CREATE:		return 0;
		case WM_ERASEBKGND: return 0;
		case WM_PAINT:
			RECT r;
			PAINTSTRUCT ps;
			if ( GetUpdateRect( hwnd, &r, FALSE ) )
			{	
				BeginPaint(hwnd, &ps) ;
				g_Plugin.HandlePaint( ps.hdc ) ;
				EndPaint(hwnd, &ps);
			}	
			return 0 ;
		case WM_ACTIVATE:
			g_Plugin.HandleActivate( wParam ) ;
			return 0 ;
		case WM_DESTROY: PostQuitMessage(0); return 0;

		case WM_KEYDOWN: // pass keyboard messages to main winamp window (for processing)
		case WM_KEYUP:
			{	
				if (( message == WM_KEYDOWN ) &&
				    ( wParam == VK_ESCAPE ) )
				{
					g_Plugin.m_bIgnoreResize = TRUE ;
					TRACE0( "User killing plugin" ) ;
					PostQuitMessage(0) ;
				}
				else if ( !g_Plugin.HandleKey( message, wParam ) )
				{
					// get this_mod from our window's user data
					// and pass the message on because the
					// plugin doesn't want to handle it
					winampVisModule *this_mod = (winampVisModule *) GetWindowLong(hwnd,GWL_USERDATA);
					PostMessage(this_mod->hwndParent,message,wParam,lParam);
				}
			}
			return 0;
		case WM_MOVE:
			{	// get config_x and config_y for configuration
				RECT r;
				GetWindowRect(hMainWnd,&r);
				config_x = r.left;
				config_y = r.top;
			}
			return 0;
        case WM_TIMER:
            if ( TIMER_ID == wParam )
            {
				KillTimer( hwnd, TIMER_ID ) ;
				TRACE0( "Automatic Plugin Death...." ) ;
				g_Plugin.m_bIgnoreResize = TRUE ;
				PostQuitMessage(0) ;
			}

		case WM_SIZE:
			{
				int width = LOWORD(lParam);
				int height = HIWORD(lParam);
				if (width && height) 
				{
					winampVisModule *this_mod = (winampVisModule *) GetWindowLong(hwnd,GWL_USERDATA);
					g_Plugin.ResizeViewport( this_mod ) ;
					g_Plugin.m_bMinimized = FALSE;
				} 
				else 
				{
					g_Plugin.m_bMinimized = TRUE;
				}
			}
			return 0;
	}
	return DefWindowProc(hwnd,message,wParam,lParam);
}

// configuration. Passed this_mod, as a "this" parameter. Allows you to make one configuration
// function that shares code for all your modules (you don't HAVE to use it though, you can make
// config1(), config2(), etc...)
void config(struct winampVisModule *this_mod)
{
	// Prepare debug output file
	g_Info.Init() ;
	DoConfigDialog( this_mod ) ;
}

// Winamp tells use to configure ourselves:
// Prepare a window, then tell g_Plugin to create
// the 3D objects.

int initD3D(struct winampVisModule *this_mod)
{
	// Prepare debug output log
	g_Info.Init() ;
	TRACE0( "Winamp telling plugin to Init" ) ;
	
	int width = WINDOW_SIZE_WIDTH ;
	int height = WIDNOW_SIZE_HEIGHT ;
	
	{	// Register our window class
		WNDCLASS wc;
		memset(&wc,0,sizeof(wc));
		wc.lpfnWndProc = WndProc;				// our window procedure
		wc.hInstance = this_mod->hDllInstance;	// hInstance of DLL
		wc.lpszClassName = szAppName;			// our window class name
	
		if (!RegisterClass(&wc)) 
		{
			// I've tryed to avoid this problem, but it might still happen -
			// last attempt crashed and didn't cleanup properly
			MessageBox(this_mod->hwndParent,"Error registering window class\nYou probaby have to restart Winamp","Plugin Error",MB_OK);
			ERRORMSG( "Fail register window class" ) ;
			return 1;
		}
	}


	hMainWnd = CreateWindowEx(
		WS_EX_APPWINDOW,
		szAppName,							// our window class name
		this_mod->description,				// use description for a window title
		WS_OVERLAPPEDWINDOW,
		config_x,config_y,					// screen position (read from config)
		width,height,						// width & height of window (need to adjust client area later)
		this_mod->hwndParent,				// parent window (winamp main window)
		NULL,								// no menu
		this_mod->hDllInstance,				// hInstance of DLL
		0); // no window creation data

	if (!hMainWnd) 
	{
		MessageBox(this_mod->hwndParent,"Error creating window","Winamp Error",MB_OK) ;
		ERRORMSG( "Failed to Create Window" ) ;
		UnregisterClass(szAppName,this_mod->hDllInstance); // unregister window class
		return 1;
	}


	SetWindowLong(hMainWnd,GWL_USERDATA,(LONG)this_mod); // set our user data to a "this" pointer

	{	// adjust size of window to make the client area exactly width x height
		RECT r;
		GetClientRect(hMainWnd,&r);
		SetWindowPos(hMainWnd,0,0,0,width*2-r.right,height*2-r.bottom,SWP_NOMOVE|SWP_NOZORDER);
	}

	// show the window
	ShowWindow(hMainWnd,SW_SHOWNORMAL);

#ifdef ESCAPE_CRAP_OUT
    // Create a timer to flip crap out
    SetTimer(hMainWnd, TIMER_ID, TIMER_RATE, NULL ) ;
#endif

	g_Plugin.m_bFirstInit = TRUE ;

	if ( g_Plugin.Init( this_mod ) == 1 )
	{
		ERRORMSG( "Plugin Init Failed - Crapping out" ) ;
		
		DestroyWindow(hMainWnd); // delete our window
		UnregisterClass(szAppName,this_mod->hDllInstance); // unregister window class
		return 1 ;
	}
	return 0 ;
}

// render function
// The MOD contains the analysis of the music for
// us to visualize
int renderD3D(struct winampVisModule *in_mod)
{
	g_Plugin.Render( in_mod ) ;	
	return 0;
}

// cleanup (opposite of init()). Destroys the window, unregisters the window class
void quitD3D(struct winampVisModule *this_mod)
{
	TRACE0( "Winamp telling Plugin to Quit" ) ;

	g_Plugin.Quit( this_mod ) ;
	g_Plugin.m_bFirstInit = FALSE ;
	DestroyWindow(hMainWnd); // delete our window
	UnregisterClass(szAppName,this_mod->hDllInstance); // unregister window class
}
