//******************************************************************************
// configdlg.cpp
//
// Andrew Skowronski, Copyright 1999
// andrew_skow@hotmail.com
//
// MFC is a lot easier than doing it from scratch like this,
// but this avoids worries about people missing the correct
// MFC DLLs on their machine.
//
// Winamp is not guaranteed to keep the same instance of the
// DLL loaded between the Configuration and the actual launch
// of the application.  We therefore need to 
// communicate all the configuration through the .ini file.
// This is good anyway, because we want to preserve user settings!
//
//******************************************************************************


#include <windowsx.h> // Wrapper macros for messages
#include <windows.h>
#include <COMMCTRL.H> // Windows messages for talking to controls
#include "3DPlugin.h"
#include "resource.h"

USE_ERROR_TRACING

//******************************************************************************
// Function Prototypes
//******************************************************************************

BOOL CALLBACK MyDialogProc( HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam ) ;

void ConfigDialogInit( HWND in_hwndDlg, 
					   struct winampVisModule * in_mod, 
					   LPPLUGINCONFIG out_Config ) ;
void ConfigDialogTerm( HWND in_hwndDlg, 
					   struct winampVisModule * in_mod, 
					   LPPLUGINCONFIG in_Config ) ;

void GetIniFilePath( struct winampVisModule * in_mod, char * out_pszFilePath ) ;

int ChooseNewColor( HWND in_hwndDlg, D3DCOLOR* io_Current) ;

BOOL GetDeviceOptions( HWND in_hwndDlg, 
					   struct winampVisModule * in_mod, 
					   LPPLUGINCONFIG out_Config ) ;

DWORD BPPToDDBD( int bpp ) ;

DWORD g_wBPP;                 // bit depth of the current display mode

#define MAX_DRIVERS 5           // maximum D3D drivers we ever expect to find

GUID g_DriverGUID[MAX_DRIVERS];     // GUIDs of the available D3D drivers
char g_DriverName[MAX_DRIVERS][50]; // names of the available D3D drivers
int  g_NumDrivers;                  // number of available D3D drivers



//******************************************************************************
// Configuration Dialog Implementation
//******************************************************************************

// Launch Configuration dialog box
void DoConfigDialog(struct winampVisModule *this_mod)
{
	DialogBoxParam( this_mod->hDllInstance, 
					MAKEINTRESOURCE( IDD_VIS_D3D_CONFIG ),
					this_mod->hwndParent,
					(DLGPROC)MyDialogProc,
					(LPARAM) this_mod ) ;
}

// Handle messages for the Dialog Box, just
// standard Windows stuff
BOOL CALLBACK MyDialogProc(  HWND hwndDlg,  // handle to dialog box
						     UINT uMsg,     
						     WPARAM wParam, 
						     LPARAM lParam  ) // second message parameter
{
	static struct winampVisModule * pMod = NULL ;
	static PluginConfig PlugConfig ;
	
	switch ( uMsg )
	{
		case WM_COMMAND:
			if ( LOWORD( wParam ) == IDOK )
			{
				ConfigDialogTerm( hwndDlg, pMod, &PlugConfig ) ;
			}
			else if ( LOWORD( wParam ) == ID_CHANGE_COLOR )
			{
				ChooseNewColor( hwndDlg, &PlugConfig.ObjectColor ) ;
			}

			break;

		case WM_INITDIALOG:

			pMod = ( struct winampVisModule * ) lParam ;
			ConfigDialogInit( hwndDlg, pMod, &PlugConfig ) ;
			return TRUE;
	}

	return 0 ; // Not processed
}

// Initialize the UI of the Configuration
// dialog and fill in the Config Structure
void ConfigDialogInit( HWND in_hwndDlg, 
					   struct winampVisModule * in_mod, 
					   LPPLUGINCONFIG out_Config )
{
	GetPluginConfig( in_mod, out_Config ) ;

	// Update the UI to match the settings

	if ( out_Config->iQuality == 0 )
	{
		HWND hQualCheckBox = GetDlgItem( in_hwndDlg, IDC_CHECK_LESS_CPU ) ;
		Button_SetCheck( hQualCheckBox, BST_CHECKED ) ;
	}

	if ( out_Config->bFullScreen )
	{
		HWND hFullScreen = GetDlgItem( in_hwndDlg, IDC_CHECK_FULLSCREEN ) ;	
		Button_SetCheck( hFullScreen, BST_CHECKED ) ;
	}

	if ( strlen( out_Config->szTextureFile ) > 0 )
	{
		HWND hTextFile = GetDlgItem( in_hwndDlg, IDC_TEXTURE_FILE ) ;
		Edit_SetText( hTextFile, out_Config->szTextureFile ) ;
	}

	HWND hTextFile = GetDlgItem( in_hwndDlg, IDC_TILE_COUNT ) ;
	char l_szVal[ 32 ] ;
	_itoa( out_Config->iTileCnt, l_szVal, 10 );
	Edit_SetText( hTextFile, l_szVal ) ;

	HWND hRough = GetDlgItem( in_hwndDlg, IDC_LANDSCAPE_ROUGHNESS ) ;
	::SendMessage(hRough, TBM_SETRANGEMIN, FALSE, (LPARAM) LANDSCAPE_ROUGH_MIN);
	::SendMessage(hRough, TBM_SETRANGEMAX, FALSE, (LPARAM) LANDSCAPE_ROUGH_MAX);

    ::SendMessage(hRough, TBM_SETPAGESIZE, 0, 1 );
	::SendMessage(hRough, TBM_SETPOS, TRUE, (LPARAM) out_Config->iLandscapeRough );

	// Record the current display BPP
    HDC hdc = GetDC( in_hwndDlg);
    g_wBPP = GetDeviceCaps( hdc, BITSPIXEL );
    ReleaseDC( in_hwndDlg, hdc );

	GetDeviceOptions( in_hwndDlg, in_mod, out_Config ) ;
}

void ConfigDialogTerm( HWND in_hwndDlg, 
					   struct winampVisModule * in_mod, 
					   LPPLUGINCONFIG in_Config )
{	
	// Read the UI state
	HWND hQualCheckBox = GetDlgItem( in_hwndDlg, IDC_CHECK_LESS_CPU ) ;
	in_Config->iQuality = !(  BST_CHECKED & Button_GetCheck( hQualCheckBox ) ) ;

	HWND hFullScreen = GetDlgItem( in_hwndDlg, IDC_CHECK_FULLSCREEN ) ;	
	in_Config->bFullScreen = (  BST_CHECKED & Button_GetCheck( hFullScreen ) ) ;

	HWND hTextFile = GetDlgItem( in_hwndDlg, IDC_TEXTURE_FILE ) ;
	Edit_GetText( hTextFile, in_Config->szTextureFile, MAX_PATH ) ;

	HWND hRoughSlider = GetDlgItem( in_hwndDlg, IDC_LANDSCAPE_ROUGHNESS ) ;
	in_Config->iLandscapeRough = ::SendMessage(hRoughSlider, TBM_GETPOS, 0, 0 ) ;

	HWND hTextCnt = GetDlgItem( in_hwndDlg, IDC_TILE_COUNT ) ;
	TCHAR l_szCount[ 32 ] ;
	Edit_GetText( hTextCnt, l_szCount, 32 ) ;
	in_Config->iTileCnt = atoi( l_szCount );

	if ( ( in_Config->iTileCnt < 1 ) || ( in_Config->iTileCnt > 1000 ) )
	{
		MessageBox( in_hwndDlg, "Invalid Tiling Count, using default value instead", "You screwed up....", MB_OK ) ;
		in_Config->iTileCnt = 1 ;
	}

	HWND hDeviceDropDown = GetDlgItem( in_hwndDlg, IDC_DEVICE_COMBO ) ;	
	int iSel = ComboBox_GetCurSel( hDeviceDropDown ) ;

	if ( ( iSel != LB_ERR ) && ( iSel < g_NumDrivers ) )
	{
		in_Config->guidDevice = g_DriverGUID[ iSel ] ;
	}

	// Save the configuration
	SavePluginConfig( in_mod, in_Config ) ;

	// Close the dialog box
	EndDialog( in_hwndDlg, TRUE ) ;
}

// Use a standard windows control to choose the color,
// most of the hastle is just converting because
// color formats
int ChooseNewColor( HWND in_hwndDlg, D3DCOLOR* io_Current)
{
    CHOOSECOLOR cc;
    COLORREF initClr;
    COLORREF aclrCust[16];
    int i;

	// We could prepare some favourite custom
	// colours
    for (i = 0; i < 16; i++)
	{
		aclrCust[i] = RGB(255, 255, 255);
	}

    initClr =
	RGB
	(   (int) (255 * D3DRMColorGetRed(*io_Current)),
	    (int) (255 * D3DRMColorGetGreen(*io_Current)),
	    (int) (255 * D3DRMColorGetBlue(*io_Current))
	);

    ZeroMemory( &cc, sizeof( CHOOSECOLOR ) );
    cc.lStructSize = sizeof( CHOOSECOLOR );
    cc.hwndOwner = in_hwndDlg;
    cc.rgbResult = initClr ;
    cc.lpCustColors = aclrCust ;
    cc.Flags = CC_RGBINIT|CC_FULLOPEN;

	// Standard Windows Color Picker
    if ( ChooseColor(&cc) )
    {	
		*io_Current = 
		D3DRMCreateColorRGB
		(	D3DVAL( GetRValue(cc.rgbResult) / D3DVAL(255.0) ),
			D3DVAL( GetGValue(cc.rgbResult) / D3DVAL(255.0) ),
			D3DVAL( GetBValue(cc.rgbResult) / D3DVAL(255.0) )
	    );

		return TRUE;
    }
    else 
	{
		return FALSE;
	}
}


//******************************************************************************
// Configuration Structure Persistence
//******************************************************************************

// Default values for the Config structure
void GetDefaultPluginConfig( LPPLUGINCONFIG out_Config )
{
	out_Config->iQuality = 1 ;

	out_Config->ObjectColor = D3DRMCreateColorRGB( D3DVAL( LAND_COLOR_R ),
												   D3DVAL( LAND_COLOR_G ),
												   D3DVAL( LAND_COLOR_B ) ) ;
	out_Config->szTextureFile[ 0 ] = 0 ;

	// I used to use GUID_NULL as the default
	// device, this worked in DX 6.0,  however
	// for some reason DX 6.1 no longer allows
	// a default Device!
	out_Config->guidDevice = IID_IDirect3DRampDevice ;

	out_Config->bFullScreen = FALSE ;

	out_Config->iResX = WINDOW_SIZE_WIDTH ;
	out_Config->iResY = WIDNOW_SIZE_HEIGHT ;

	out_Config->iTileCnt = 1 ;
	out_Config->iLandscapeRough = LANDSCAPE_ROUGH ;
}

// Load the configuration out of the ini file.
// If the settings have not been saved then default values
// are used
void GetPluginConfig( struct winampVisModule * in_mod, LPPLUGINCONFIG out_Config )
{
	// Initialize with default values in
	// case nothing has been persisted
	GetDefaultPluginConfig( out_Config ) ;
	
	char szInitFile[MAX_PATH];
	GetIniFilePath( in_mod, szInitFile ) ;

	out_Config->iQuality = GetPrivateProfileInt( in_mod->description, 
												 "Quality", 
												 out_Config->iQuality,	// Default
												 szInitFile );

	out_Config->ObjectColor = GetPrivateProfileInt( in_mod->description, 
												 "ObjectColor", 
												 out_Config->ObjectColor,	// Default
												 szInitFile );

	out_Config->iLandscapeRough = GetPrivateProfileInt( in_mod->description, 
												 "Roughness", 
												 out_Config->iLandscapeRough, // Default
												 szInitFile );

	out_Config->iTileCnt = GetPrivateProfileInt( in_mod->description, 
												 "Tile Count", 
												 out_Config->iTileCnt,	// Default
												 szInitFile );

	
	out_Config->bFullScreen = GetPrivateProfileInt( in_mod->description, 
												 "Full Screen", 
												 out_Config->bFullScreen,	// Default
												 szInitFile );


	GetPrivateProfileString( in_mod->description, 
							 "TextureFile", 
							 "", 
							 out_Config->szTextureFile, 
							 MAX_PATH, 
							 szInitFile ) ;

	// All this trouble just to load the Device GUID
	// welcome to the wonderful world of char, WCHAR, and OLECHAR
	char szGuid[ 64 ] ;
	WCHAR wszGuid[ 64 ] ;

	GetPrivateProfileString( in_mod->description,
							 "Device", 
							 "", 
							 szGuid, 
							 64, 
							 szInitFile ) ;

	if ( szGuid[ 0 ] != 0 )
	{
		MultiByteToWideChar( CP_ACP, 0, szGuid, -1, wszGuid, 64 ) ;

		CLSIDFromString( wszGuid, &out_Config->guidDevice ) ;
	}
}

// Save the Config Structure into the ini file
void SavePluginConfig( struct winampVisModule * in_mod, LPPLUGINCONFIG in_Config )
{
	char szInitFile[MAX_PATH];
	GetIniFilePath( in_mod, szInitFile ) ;

	char szString[32];

	wsprintf( szString,"%d",in_Config->iQuality ) ;
	WritePrivateProfileString( in_mod->description,"Quality",szString,szInitFile) ;

	wsprintf( szString,"%d",in_Config->ObjectColor ) ;
	WritePrivateProfileString( in_mod->description,"ObjectColor",szString,szInitFile) ;

	WritePrivateProfileString( in_mod->description,"TextureFile",in_Config->szTextureFile,szInitFile) ;

	WCHAR wszGuid[ 64 ] ;
	char  szGuid[ 64 ] ;
	StringFromGUID2( in_Config->guidDevice, wszGuid, 64 ) ;

	WideCharToMultiByte( CP_ACP, 0, wszGuid, -1, szGuid, 64, NULL, NULL ) ;

	WritePrivateProfileString( in_mod->description,"Device",szGuid, szInitFile) ;

	wsprintf( szString,"%d",in_Config->iTileCnt ) ;
	WritePrivateProfileString( in_mod->description,"Tile Count",szString,szInitFile) ;

	wsprintf( szString,"%d",in_Config->iLandscapeRough ) ;
	WritePrivateProfileString( in_mod->description,"Roughness",szString,szInitFile) ;

	wsprintf( szString,"%d",in_Config->bFullScreen ) ;
	WritePrivateProfileString( in_mod->description,"Full Screen",szString,szInitFile) ;
}

// Determines the path of the configuration file to use
void GetIniFilePath( struct winampVisModule * in_mod, char * out_pszFilePath )
{
	char szDirectory[ MAX_PATH ] ;
	char szModulePath[ MAX_PATH ] ;
	char szDrive[ 5 ] ;

	out_pszFilePath[0] = 0 ;

	GetModuleFileName( in_mod->hDllInstance, szModulePath, MAX_PATH ) ;	
	_splitpath( szModulePath, szDrive, szDirectory, NULL, NULL ) ;

	strcat( out_pszFilePath, szDrive ) ;
	strcat( out_pszFilePath, szDirectory ) ;
	strcat( out_pszFilePath, "plugin.ini" ) ;
}

// Argh, what a waste of code...
DWORD BPPToDDBD( int bpp )
{
    switch( bpp )
	{
        case 1:
            return DDBD_1;
        case 2:
            return DDBD_2;
        case 4:
            return DDBD_4;
        case 8:
            return DDBD_8;
        case 16:
            return DDBD_16;
        case 24:
            return DDBD_24;
        case 32:
            return DDBD_32;
        default:
            return 0;
    }
}


HRESULT WINAPI enumDeviceFunc( LPGUID lpGuid, LPSTR lpDeviceDescription,
							   LPSTR lpDeviceName, LPD3DDEVICEDESC lpHWDesc, 
							   LPD3DDEVICEDESC lpHELDesc, LPVOID lpContext )
{
    LPD3DDEVICEDESC lpDesc;

	TRACE0( "\n-----Examining Device-----" ) ;
	TRACE0( lpDeviceName ) ;

	// Decide which device description we should consult
    lpDesc = lpHWDesc->dcmColorModel ? lpHWDesc : lpHELDesc;

	// If this driver cannot render in the current display bit depth skip
    // it and continue with the enumeration.
	//
	// NOTE: If we run in FULL screen mode we can ignore
	// this restriction, right?
    if (!(lpDesc->dwDeviceRenderBitDepth & BPPToDDBD(g_wBPP)))
        return D3DENUMRET_OK;

	// Record this driver's info
    memcpy(&g_DriverGUID[g_NumDrivers], lpGuid, sizeof(GUID));
    lstrcpy(&g_DriverName[g_NumDrivers][0], lpDeviceName);

	// Some interesting information for debugging
	if ( lpDesc == lpHWDesc )
	{
		TRACE0( "Found Hardware Driver\n" ) ;
	}
	else
	{
		TRACE0( "Found Software Driver\n" ) ;
	}

	if ( lpDesc->dcmColorModel & D3DCOLOR_MONO )
	{
		TRACE0( "Device supports Mono Lighting Model" ) ;
	}
	else
	{
		TRACE0( "Device supports RGB Lighting Model" ) ;
	}

    g_NumDrivers++;

    if (g_NumDrivers == MAX_DRIVERS)
	{
        return D3DENUMRET_CANCEL;
	}

    return D3DENUMRET_OK;
}

BOOL GetDeviceOptions( HWND in_hwndDlg, 
					   struct winampVisModule * in_mod, 
					   LPPLUGINCONFIG out_Config )
{
    LPDIRECTDRAW pDD = NULL ;
    LPDIRECT3D   pD3D = NULL ;

	TRACE0( "Determining Available Direct 3D Devices" ) ;

	// TODO: While we are at it we should verify that
	// they have at least Direct X 6...

    // Create a DirectDraw object and query for the Direct3D interface to use
    // to enumerate the drivers.
    if( FAILED( DirectDrawCreate( NULL, &pDD, NULL ) ) )
	{
        ERRORMSG( "Creation of DirectDraw HEL failed." );
        return FALSE;
    }

    if( FAILED( pDD->QueryInterface( IID_IDirect3D, (VOID**)&pD3D ) ) )
	{
        ERRORMSG( "Creation of Direct3D interface failed." );
        pDD->Release();
        return FALSE;    
	}

	g_NumDrivers = 0 ;

/*
	// Doesn't work anymore in DX 6.1, I'm leaving this
	// here for a while in case this is a bug that they
	// will fix

    // Add the default driver
	g_DriverGUID[ 0 ] = GUID_NULL ;
	strcpy( g_DriverName[ 0 ], "Default Driver" ) ;
*/

	// Enumerate the drivers
    if( FAILED( pD3D->EnumDevices( enumDeviceFunc, NULL ) ) )
	{
        ERRORMSG("Enumeration of drivers failed." );
        return FALSE;
    }

    // Close Direct 3D Again
	RELEASE( pD3D ) ;
	RELEASE( pDD ) ;

	// Add the driver names
	HWND hDeviceDropDown = GetDlgItem( in_hwndDlg, IDC_DEVICE_COMBO ) ;

    for( int i = 0; i < g_NumDrivers; i++ )
	{
		ComboBox_AddString( hDeviceDropDown, g_DriverName[i] ) ;
	}
	
	ComboBox_SetCurSel( hDeviceDropDown, 0 ) ;

	// Try to initially select the persisted value
	for( i = 0 ; i < g_NumDrivers ; i++ )
	{
		if ( g_DriverGUID[i] == out_Config->guidDevice )
		{
			ComboBox_SetCurSel( hDeviceDropDown, i ) ;
			break ;
		}
	}

    return TRUE ;
}
