//******************************************************************************
// light.cpp
// Implementation of CLight
//
// Andrew Skowronski, Copyright 1999
// andrew_skow@hotmail.com
// 
//******************************************************************************

#include "3DPlugin.h"

CLight::CLight()
	: m_pLightFrame( NULL )
	, m_pLight( NULL )
	, m_bActive( FALSE )
{

}

CLight::~CLight()
{
	RELEASE( m_pLight ) ;
	RELEASE( m_pLightFrame ) ;
}

// Create the light and the frame for the light,
// but don't turn it on yet.

HRESULT CLight::Init( D3DRMLIGHTTYPE in_Type,
	       float in_Brightness,
		   float in_xPos,
		   float in_yPos,
		   float in_zPos,
		   float in_xRot,
		   float in_yRot,
		   float in_zRot,
		   float in_RotAmount,
		   LPDIRECT3DRMFRAME3 in_pRelativeFrame,
		   LPDIRECT3DRM3 in_pDRM3 )
{
	// Validate State
	if ( m_bActive )
	{
		return E_UNEXPECTED ;
	}

	if ( in_pRelativeFrame == NULL )
	{
		return E_INVALIDARG ;
	}

	if ( in_pDRM3 == NULL )
	{
		return E_INVALIDARG ;
	}

	HRESULT hr ;

	hr = in_pDRM3->CreateLightRGB( in_Type, 
                                   D3DVAL(in_Brightness), 
								   D3DVAL(in_Brightness), 
								   D3DVAL(in_Brightness), 
								   &m_pLight ) ;
	
	if( FAILED( hr ) ) return hr ;
	
	//
	// Create Frames that will contain the lights
	//

	hr = in_pDRM3->CreateFrame( in_pRelativeFrame, 
							    &m_pLightFrame ) ;
	if( FAILED( hr ) ) return hr ;

	//
	// Positions the lights in the scene
	//

    hr = m_pLightFrame->SetPosition(in_pRelativeFrame, 
                                    D3DVAL(in_xPos), 
									D3DVAL(in_yPos), 
									D3DVAL(in_zPos) ) ;
	
	if( FAILED( hr ) ) return hr ;
	

	if ( in_xRot != 0.0 || in_yRot != 0.0 || in_zRot != 0.0 )
	{
		if ( in_RotAmount == 0.0 )
		{
			return E_INVALIDARG ;
		}

		hr = m_pLightFrame->SetRotation( in_pRelativeFrame,
                                         D3DVAL(in_xRot), 
									     D3DVAL(in_yRot), 
									     D3DVAL(in_zRot),
									     D3DVAL(in_RotAmount) ) ;
	}

	return hr ;
}

// Turns the light on or off 
HRESULT CLight::ToggleActive()
{
	HRESULT hr ;

	if (( m_pLightFrame == NULL ) ||
		( m_pLight      == NULL ) )
	{
		// Init Wasn't Called
		return E_UNEXPECTED ;
	}

	if ( !m_bActive )
	{
		hr = m_pLightFrame->AddLight( m_pLight ) ;
	}
	else
	{
		hr = m_pLightFrame->DeleteLight( m_pLight ) ;
	}

	m_bActive = !m_bActive ;

	return hr ;
}